@@ -1,5 +1,6 @@
 import typing
 import inspect
+import pandas
 
 
 class LoadStrategy(object):
@@ -8,7 +9,7 @@ class LoadStrategy(object):
     """
     def __init__(self, format: str=None, read_args: dict=None, use_dask: bool=False, read_method: typing.Callable=None,
                  uri: str=None, open_args: dict=None, opener: typing.Callable=None,
-                 predefined=None, sql: str=None):
+                 predefined=None, sql: str=None, sql_parameters: list=None):
         """
         :param format:      File format, as a file extension, i.e. "csv", etc.
         :param read_args:   Arguments to send to the detected read_method.
@@ -19,6 +20,7 @@ class LoadStrategy(object):
         :param opener:      Method that opens a stream to read this data.
         :param predefined:  Predefined DataFrame to use instead of anything else.
         :param sql:         SQL, if data is coming from a database.
+        :param sql_parameters: Optional parameters for the SQL, in case it is parameterized.
         """
         self.format = format
         self.read_args = read_args
@@ -29,6 +31,7 @@ class LoadStrategy(object):
         self.opener = opener
         self.predefined = predefined
         self.sql = sql
+        self.sql_parameters = sql_parameters
 
     @staticmethod
     def method_name(method: typing.Callable):
@@ -119,7 +122,10 @@ class LoadStrategy(object):
             connect_code = f"{method_o}({self.repr_kwargs(self.open_args, comma_before=False)})"
             # TODO: here, 'read_method' is assumed to be a self-contained local utility method - some kind of verification should happen
             code += inspect.getsource(self.read_method) + "\n"
-            code += f"{result_var} = {self.read_method.__name__}({connect_code})\n"
+            read_code = f", sql={repr(self.sql)}"
+            if self.sql_parameters:
+                read_code += f", sql_parameters={repr(self.sql_parameters)}"
+            code += f"{result_var} = {self.read_method.__name__}({connect_code}{read_code})\n"
         else:
             read_args = self.read_args
             mod_o, method_o = self.method_name(self.opener)
@@ -230,7 +236,7 @@ class ApplySubindexing(object):
             return self.parent.tell()
 
 
-def load_dataframe_from_sql(db_conn, sql: str, use_dask: int=0):
+def load_dataframe_from_sql(db_conn, sql: str, parameters=None, use_dask: int=0):
     """
     Given a database connection, load a dataframe.
 
@@ -240,7 +246,7 @@ def load_dataframe_from_sql(db_conn, sql: str, use_dask: int=0):
     try:
         import pandas
         c = db_conn.cursor()
-        c.execute(sql)
+        c.execute(sql, parameters)
         col_names = [col[0] for col in c.description]
         df = pandas.DataFrame(data=c, columns=col_names)
         # TODO here we are, dutifully moving a too-big dataframe to dask, is it worth it?
@@ -253,3 +259,40 @@ def load_dataframe_from_sql(db_conn, sql: str, use_dask: int=0):
         return df
     finally:
         db_conn.close()
+
+
+def coerce_to_dataframe(content):
+    """
+    Turn things into dataframes.
+    """
+    if hasattr(content, "columns") and hasattr(content, "itertuples"):
+        return content
+    elif hasattr(content, "get") and content.get("data"):
+        # arguments to pandas.DataFrame constructor, i.e. data=..., columns=...
+        return pandas.DataFrame(**content)
+    elif hasattr(content, "__iter__"):
+        content = pandas.DataFrame(content)
+        if len(content.columns) and not isinstance(content.columns[0], str):
+            # numeric column names export ambiguously to CSV
+            # TODO align this convention with pandas methods which generate (maybe) "Col_1", etc.
+            content.columns = [f"col_{n}" for n in range(1, len(content.columns)+1)]
+        return content
+    else:
+        raise NotImplementedError(f"cannot convert {type(content)} to dataframe")
+
+
+def convertible_to_dataframe(content):
+    """
+    Detect things that can be turned into dataframes.
+    """
+    if hasattr(content, "columns") and hasattr(content, "itertuples"):
+        # a dataframe
+        return True
+    if hasattr(content, "get") and content.get("data"):
+        # {}-like arguments to pandas.DataFrame constructor, i.e. data=..., columns=...
+        return True
+    if hasattr(content, "__iter__") and not isinstance(content, str):
+        # a sequence of rows, or a {} with named columns
+        # TODO verify compatibility with pandas.DataFrame() constructor more extensively here
+        return True
+    return False
